import { Context, Middleware as IMiddleware } from 'koa'
import * as request from 'superagent'
import { diff } from 'deep-object-diff'
import * as Debug from 'debug'
import { getSession, createSession, updateSession } from './requests'

const log = Debug('middleware:session')

declare module 'koa' {
  interface User {
    uuid: string
    name: string
    github?: string
    youtube?: string
    twitch?: string
    points: number
  }

  interface Session {
    id: string
    maxAge: number
    user?: User
    isLoggedIn: boolean
    identifier: string
    [key: string]: any
  }

  interface Context {
    session: Session
  }
}

export interface ISessionMiddlewareOptions {
  url?: string
  domain?: string
}

const THIRTY_DAYS = 1000 * 60 * 60 * 24 * 30

const defaultOptions: ISessionMiddlewareOptions = {
  url: 'localhost:3100',
  domain: 'sr.dev',
}

export function Session(optionsIn?: ISessionMiddlewareOptions): IMiddleware {
  const options = Object.assign({}, defaultOptions, optionsIn)
  const doPrintDebug = /session-middleware/.test(process.env.DEBUG || '')
  // tslint:disable-next-line:no-console
  const print = (str: string) => doPrintDebug && console.log(str)

  // Ping auth service
  request
    .get(`${options.url}`)
    .on('error', () => {
      throw new Error("Couldn't reach auth service.")
    })
    .then(response => {
      if (response.text !== 'What are you doing here?') {
        throw new Error(
          `Auth service responded incorrectly (${response.text}).`
        )
      }

      print('Auth service responded correctly')
    })

  return async function Middleware(ctx: Context, next: () => Promise<any>) {
    const sid = ctx.cookies.get('sid')
    let session: any

    if (sid) {
      log('Found a SID')
      session = await getSession(options.url as string, sid)
    } else {
      log("Couldn't find a SID")
      session = await createSession(options.url as string)
    }

    ctx.session = {
      id: session.id,
      maxAge: session.maxAge,
      user: session.user,
      isLoggedIn: !!session.user,
      identifier: session.user ? session.user.uuid : session.id,

      ...session.session,
    }

    ctx.cookies.set('sid', session.id, {
      maxAge: THIRTY_DAYS,
      domain: '.' + options.domain,
      secure: process.env.NODE_ENV === 'production',
    })

    await next()

    const { id, maxAge, user, isLoggedIn, identifier, ...rest } = ctx.session
    const ctxSession = { id, maxAge, user, session: rest }
    const difference = diff(session, ctxSession)

    if (Object.keys(difference).length > 0) {
      log('Found difference in session!')
      log(difference)
      await updateSession(options.url as string, id, diff(session, ctxSession))
    }
  }
}
